// ***********************************************************
// Project: Magnetometer Sensor LIS3MDL
// Author: Dr. Simon Schrdle
//
// Started: 5.12.2015
// ***********************************************************


#define F_CPU 16000000UL  // 16 MHz


#include <avr\io.h>
#include <avr\interrupt.h>
#include <avr\signal.h>
#include <stdlib.h>
#include <avr/wdt.h>
#include "usbdrv/usbdrv.h"
#include <util/delay.h>

//lis3mdl lines
#define MAG_CS    0x04 //to mag
#define MAG_SPC   0x02 //to mag
#define MAG_SDI   0x01 //to mag
#define MAG_SDO   0x08 //from mag


#define MAG_PORT  PORTC
#define MAG_PIN   PINC
#define MAG_DDR	DDRC

#define USB_INFOMSG 	0


#define USB_MAGWHOAMI 	5 //should return 0x3d mag device id

//write to device commands
#define USB_SETR1 	10
#define USB_MAGFAST 	11 //fast meas mode

#define USB_MAGFAST2 	12 //fast meas mode self test

//read from device commands
#define USB_GETX 		20
#define USB_GETY 		21
#define USB_GETZ 		22
#define USB_GETXYZ 	25
#define USB_GETXYZF 	26 //fast read method
#define USB_GETXYZFA 	27 //fast read method avg.

void wait_10k(void) {  /*wait 10 k cycles*/
uint16_t counter;

  counter=10000;
  do {
  } while (--counter);
}

void wait_100(void) {  /*wait 100 cycles*/
uint16_t counter;

  counter=100;
  do {
  } while (--counter);
}


//mag functions

void mag_init(void) {
 MAG_DDR  |= MAG_CS|MAG_SDI|MAG_SPC; //output from AVR
 MAG_DDR  &= ~(MAG_SDO); //inputs to AVR

 MAG_PORT &= ~(MAG_SDI|MAG_SPC|MAG_SDO); //Auf LOW
 MAG_PORT |= MAG_CS; //CE High
}


void mag_write(uint8_t magaddr,uint8_t magcmd) { /*write 8 bits to addr*/
  uint8_t i;
  uint16_t bdata;

  MAG_PORT &= ~MAG_SDI;
  MAG_PORT |= MAG_SPC;
  asm("nop");
  asm("nop");
  asm("nop");

  MAG_PORT &= ~MAG_CS; //CS Low

  bdata=((magaddr&0x3f)<<8)|magcmd;

  i=16;  // send 1 word

  do {
    MAG_PORT &= ~MAG_SPC;
    if (bdata&0x8000) {MAG_PORT |= MAG_SDI;} else {MAG_PORT &= ~MAG_SDI;}
    bdata<<=1;
    asm("nop");
    asm("nop");
    asm("nop");
    MAG_PORT |= MAG_SPC; // clock cycle, insert delay if necessary, depends on CPU frequency
    asm("nop");
    asm("nop");
    asm("nop");
   } while (--i);


  MAG_PORT |= MAG_CS; //CS High
  MAG_PORT &= ~MAG_SDI; //data line auf low
}


uint8_t mag_data, mag_fast;

void mag_read(uint8_t magadr) { /*read 8 bits*/
  uint8_t i;
  uint8_t bdata,mdata;

//  MAG_PORT &= ~MAG_SDO;
  MAG_PORT |= MAG_SPC;
  asm("nop");
  asm("nop");
  asm("nop");

  MAG_PORT &= ~MAG_CS; //CS Low

  bdata=(magadr&0x3f)|0x80; //read data, 6 bit address, no reg increment

  i=8;  // send 1 byte

  do {
    MAG_PORT &= ~MAG_SPC;
    if (bdata&0x80) {MAG_PORT |= MAG_SDI;} else {MAG_PORT &= ~MAG_SDI;}
    bdata<<=1;
    asm("nop");
    asm("nop");
    asm("nop");
    MAG_PORT |= MAG_SPC; // clock cycle, insert delay if necessary, depends on CPU frequency
    asm("nop");
    asm("nop");
    asm("nop");
   } while (--i);
  i=8;  // send 1 byte

  mdata=0;
  do {
    MAG_PORT &= ~MAG_SPC;
    mdata<<=1;
    asm("nop");
    asm("nop");
    asm("nop");
    MAG_PORT |= MAG_SPC; // clock cycle, insert delay if necessary, depends on CPU frequency
    asm("nop");
    asm("nop");
    asm("nop");
    if (MAG_PIN&MAG_SDO) {mdata|=0x01;}

   } while (--i);


  MAG_PORT |= MAG_CS; //CS High
  MAG_PORT &= ~MAG_SDI; //data line auf low

  mag_data=mdata;
}


// END mag Functions


uint16_t setr1_value, mag_x,mag_y,mag_z;

void mag_setup_cont(void)
{
   mag_write(0x21,0x40); //12 gaus rang
//   mag_write(0x20,0xfc); //UHP mode on x and y, 80 hz, temp on

   mag_write(0x20,0xe4); //UHP mode on x and y, 1.25 hz, temp on

   mag_write(0x23,0x0c); //UHP mode on z
   mag_write(0x22,0x00); //run continuously
}


void mag_setup_cont_fast(void)
{
   mag_write(0x21,0x40); //12 gauss rang
//    mag_write(0x21,0x00); //4 gauss rang

//   mag_write(0x20,0xfc); //UHP mode on x and y, 80 hz, temp on
   mag_write(0x20,0xfe); //UHP mode on x and y, 155 hz, temp on

//   mag_write(0x20,0xe4); //UHP mode on x and y, 1.25 hz, temp on

   mag_write(0x23,0x0c); //UHP mode on z
   mag_write(0x22,0x00); //run continuously
}

void mag_setup_cont_fast2(void)  //self test on
{
   mag_write(0x21,0x40); //12 gauss rang
//    mag_write(0x21,0x00); //4 gauss rang

//   mag_write(0x20,0xfc); //UHP mode on x and y, 80 hz, temp on
   mag_write(0x20,0xff); //UHP mode on x and y, 155 hz, temp on   self test on

//   mag_write(0x20,0xe4); //UHP mode on x and y, 1.25 hz, temp on

   mag_write(0x23,0x0c); //UHP mode on z
   mag_write(0x22,0x00); //run continuously
}



uint16_t xavg,yavg,zavg;
int32_t xsum,ysum,zsum,sumdiv;
int16_t xv,yv,zv;
uint8_t avgc;

void mag_update(void)
{

   if (mag_fast)
   {
       wait_100();
       mag_read(0x27);

       if (mag_data&0x08) //net dataset available
       {
        mag_read(0x29);
        mag_x=mag_data;
        mag_x<<=8;
        mag_read(0x28); //read mag x data
        mag_x|=mag_data;

        mag_read(0x2b);
        mag_y=mag_data;
        mag_y<<=8;
        mag_read(0x2a); //read mag y data
        mag_y|=mag_data;

        mag_read(0x2d);
        mag_z=mag_data;
        mag_z<<=8;
        mag_read(0x2c); //read mag z data
        mag_z|=mag_data;

        xv=mag_x;
        yv=mag_y;
        zv=mag_z;

        xsum+=xv;
        ysum+=yv;
        zsum+=zv;
        avgc++;

        if (avgc==155)
        {
          avgc=0;
          sumdiv=155;

          xsum=xsum/sumdiv;
          ysum=ysum/sumdiv;
			 zsum=zsum/sumdiv;
			
			 xavg=xsum;
          yavg=ysum;
          zavg=zsum;

          xsum=0;
          ysum=0;
          zsum=0;
        }

       }

   }
}


static uchar replyBuf[16] = "TWSMAG R01 SN01";

static uchar usbd_val[3] = "QX";
static uchar usbd_val3[7] = "XXYYZZ";


// this gets called when custom control message is received
USB_PUBLIC uchar usbFunctionSetup(uchar data[8]) {
    usbRequest_t *rq = (void *)data; // cast data to correct type

    switch(rq->bRequest) { // custom command is in the bRequest field

    case USB_SETR1:
        setr1_value = rq->wValue.word;
        return 0;

    case USB_MAGFAST:
        mag_fast = rq->wValue.word;
        if (mag_fast) {mag_setup_cont_fast();} else
         {      mag_setup_cont();}
        return 0;

    case USB_MAGFAST2:
        mag_fast = rq->wValue.word;
        if (mag_fast) {mag_setup_cont_fast2();} else
         {      mag_setup_cont();}
        return 0;


    case USB_GETX:
        mag_read(0x29);
        mag_x=mag_data;
        mag_x<<=8;
        mag_read(0x28); //read mag x data
        mag_x|=mag_data;

        usbd_val[0]=mag_x>>8;
        usbd_val[1]=mag_x&0x00ff;
        usbMsgPtr = (uint16_t) ((uint8_t*)usbd_val);
        return sizeof(usbd_val);

    case USB_GETY:
        mag_read(0x2b);
        mag_y=mag_data;
        mag_y<<=8;
        mag_read(0x2a); //read mag y data
        mag_y|=mag_data;

        usbd_val[0]=mag_y>>8;
        usbd_val[1]=mag_y&0x00ff;
        usbMsgPtr = (uint16_t) ((uint8_t*)usbd_val);
        return sizeof(usbd_val);

   case USB_GETZ:
        mag_read(0x2d);
        mag_z=mag_data;
        mag_z<<=8;
        mag_read(0x2c); //read mag z data
        mag_z|=mag_data;

        usbd_val[0]=mag_z>>8;
        usbd_val[1]=mag_z&0x00ff;
        usbMsgPtr = (uint16_t) ((uint8_t*)usbd_val);
        return sizeof(usbd_val);

   case USB_GETXYZ:
        mag_read(0x29);
        mag_x=mag_data;
        mag_x<<=8;
        mag_read(0x28); //read mag x data
        mag_x|=mag_data;

        mag_read(0x2b);
        mag_y=mag_data;
        mag_y<<=8;
        mag_read(0x2a); //read mag y data
        mag_y|=mag_data;

        mag_read(0x2d);
        mag_z=mag_data;
        mag_z<<=8;
        mag_read(0x2c); //read mag z data
        mag_z|=mag_data;

        usbd_val3[0]=mag_x>>8;
        usbd_val3[1]=mag_x&0x00ff;
        usbd_val3[2]=mag_y>>8;
        usbd_val3[3]=mag_y&0x00ff;
        usbd_val3[4]=mag_z>>8;
        usbd_val3[5]=mag_z&0x00ff;

        usbMsgPtr = (uint16_t) ((uint8_t*)usbd_val3);
        return sizeof(usbd_val3);

    case USB_GETXYZF:  //fast value return
        usbd_val3[0]=mag_x>>8;
        usbd_val3[1]=mag_x&0x00ff;
        usbd_val3[2]=mag_y>>8;
        usbd_val3[3]=mag_y&0x00ff;
        usbd_val3[4]=mag_z>>8;
        usbd_val3[5]=mag_z&0x00ff;

        usbMsgPtr = (uint16_t) ((uint8_t*)usbd_val3);
        return sizeof(usbd_val3);

    case USB_GETXYZFA:  //fast value return       avg.
        usbd_val3[0]=xavg>>8;
        usbd_val3[1]=xavg&0x00ff;
        usbd_val3[2]=yavg>>8;
        usbd_val3[3]=yavg&0x00ff;
        usbd_val3[4]=zavg>>8;
        usbd_val3[5]=zavg&0x00ff;

        usbMsgPtr = (uint16_t) ((uint8_t*)usbd_val3);
        return sizeof(usbd_val3);

   case USB_MAGWHOAMI:
        mag_read(0x0f); //read mag id
        usbd_val[0]=0;
        usbd_val[1]=mag_data;
        usbMsgPtr = (uint16_t) ((uint8_t*)usbd_val);
        return sizeof(usbd_val);

    case USB_INFOMSG: // send data to PC
        usbMsgPtr =(uint16_t) ((uint8_t*) replyBuf);
        return sizeof(replyBuf);

    }
    return 0; // should not get here

}


int main(void) {
   uint8_t i;

   wdt_enable(WDTO_1S); // enable 1s watchdog timer
   usbInit();
   usbDeviceDisconnect(); // enforce re-enumeration

    for(i = 0; i<250; i++) { // wait 500 ms
        wdt_reset(); // keep the watchdog happy
        _delay_ms(2);
    }
   usbDeviceConnect();

   sei(); // Enable interrupts after re-enumeration

   mag_init();
   mag_setup_cont();


   mag_fast=0;

   mag_x=1234;
   mag_y=2345;
   mag_z=3456;


   //mag_write(0x0f00);

   do {
       wdt_reset(); // keep the watchdog happy
       usbPoll();

		 mag_update();

  } while (1);

 return 0;

}
















